"""Arctans."""

from arctans.arctans import arccotan, arctan
from arctans.numbers import Integer, Rational, GaussianInteger, GaussianRational, j, one, zero
from arctans.primes import is_irreducible
from arctans.reduction import reduce, convert_rational
from arctans.generation import generate

__version__ = "1.0.0"
